import functools


def stringify_first_arg(func):
    """Convert the first argument to a string."""

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        first_arg_string = str(args[0])
        other_args = args[1:]
        return func(first_arg_string, *other_args, **kwargs)

    return wrapper


def enumerate_human_readable_range(range_string, range_split_string: str = '-'):
    """Enumerate the range specified by the string. For example, `1-3` returns `[1, 2, 3]`."""
    range_sections = range_string.split(range_split_string)
    error_message = (
        'The enumerate_range function expects a string with two integers separated '
        + 'by the character specified by the `range_split_string` argument which can be passed '
        + 'into the enumerate_range function.'
    )

    if len(range_sections) != 2:
        raise ValueError(error_message)

    try:
        range_start = int(range_sections[0].strip())
        range_end = int(range_sections[1].strip())
    except ValueError:
        raise ValueError(error_message)
    else:
        return [i for i in range(range_start, range_end + 1)]



import pytest

from d8s_asns import asn_name  # asn_whois,
from d8s_asns import (
    asn_adjacent_asns,
    asn_announced_prefixes,
    asn_is_private,
    asn_number,
    asn_standardize,
    asns,
    asns_find,
    asns_private_numbers,
    asns_private_ranges,
    standardize_asn_decorator,
)
from d8s_asns.asns import _cidr_report_org_asn_format


def test__cidr_report_org_asn_format_docs_1():
    assert _cidr_report_org_asn_format('asn1234') == 'asn1234'


def test_asn_adjacent_asns_docs_1():
    assert tuple(asn_adjacent_asns('AS48085')) == ('AS6702', 'AS3326', 'AS1')


def test_asn_announced_prefixes_docs_1():
    assert tuple(asn_announced_prefixes('AS48085')) == ('91.210.36.0/22',)
    assert tuple(asn_announced_prefixes('AS29132')) == (
        '62.220.244.0/22',
        '185.213.224.0/22',
        '212.94.64.0/20',
        '212.94.80.0/23',
        '212.94.88.0/24',
        '212.94.89.0/24',
        '212.94.90.0/23',
        '212.94.92.0/22',
    )


def test_asn_is_private_docs_1():
    assert asn_is_private('0')
    assert asn_is_private('23456')
    assert asn_is_private(23456)
    assert asn_is_private('64496')
    assert asn_is_private('64500')
    assert asn_is_private('64511')
    assert asn_is_private('4294967295')
    assert asn_is_private('64496')
    assert asn_is_private('ASN64496')
    assert asn_is_private('ASN 64496')
    assert asn_is_private('AS64496')
    assert asn_is_private('AS 64496')

    assert not asn_is_private('AS 1234')


def test_asn_name_docs_1():
    assert asn_name('AS1') == 'LVLT-1, US'
    assert asn_name('ASN1') == 'LVLT-1, US'
    assert asn_name('AS 1') == 'LVLT-1, US'


def test_asn_number_docs_1():
    assert asn_number('64496') == 64496
    assert asn_number('ASN64496') == 64496
    assert asn_number('ASN 64496') == 64496
    assert asn_number('AS64496') == 64496
    assert asn_number('AS 64496') == 64496


def test_asn_standardize_docs_1():
    assert asn_standardize('1234') == 'ASN1234'
    assert asn_standardize('AS1234') == 'ASN1234'
    assert asn_standardize('AS 1234') == 'ASN1234'
    assert asn_standardize('ASN1234') == 'ASN1234'
    assert asn_standardize('ASN 1234') == 'ASN1234'
    assert asn_standardize('1234') == 'ASN1234'
    assert asn_standardize(1234) == 'ASN1234'
    assert asn_standardize('foo') is None


def test_asns_private_ranges_docs_1():
    assert tuple(asns_private_ranges()) == (
        {'AS Number': '0', 'Reason for Reservation': 'Reserved by [RFC7607]', 'Reference': '[RFC7607]'},
        {
            'AS Number': '112',
            'Reason for Reservation': 'Used by the AS112 project to sink misdirected DNS queries; see [RFC7534]',
            'Reference': '[RFC7534]',
        },
        {'AS Number': '23456', 'Reason for Reservation': 'AS_TRANS; reserved by [RFC6793]', 'Reference': '[RFC6793]'},
        {
            'AS Number': '64496-64511',
            'Reason for Reservation': 'For documentation and sample code; reserved by [RFC5398]',
            'Reference': '[RFC5398]',
        },
        {
            'AS Number': '64512-65534',
            'Reason for Reservation': 'For private use; reserved by [RFC6996]',
            'Reference': '[RFC6996]',
        },
        {'AS Number': '65535', 'Reason for Reservation': 'Reserved by [RFC7300]', 'Reference': '[RFC7300]'},
        {
            'AS Number': '65536-65551',
            'Reason for Reservation': 'For documentation and sample code; reserved by [RFC5398]',
            'Reference': '[RFC5398]',
        },
        {
            'AS Number': '4200000000-4294967294',
            'Reason for Reservation': 'For private use; reserved by [RFC6996]',
            'Reference': '[RFC6996]',
        },
        {'AS Number': '4294967295', 'Reason for Reservation': 'Reserved by [RFC7300]', 'Reference': '[RFC7300]'},
    )


# def test_asn_whois_docs_1():
#     results = asn_whois('AS209711')
#     print('results {}'.format(results))
#     assert results.startswith('whois: 399260')
#     assert results.endswith(')')
#     assert '% This query was served by the RIPE Database Query Service version ' in results
#     assert 'IANA has recorded AS209711 as orig' in results
#     assert 'organisation:   ORG-MBVT3-RIPE' in results

#     results = asn_whois('209711')
#     assert results.startswith('whois: 399260')
#     assert results.endswith(')')
#     assert '% This query was served by the RIPE Database Query Service version ' in results
#     assert 'IANA has recorded AS209711 as orig' in results
#     assert 'organisation:   ORG-MBVT3-RIPE' in results


def test_asns_find_docs_1():
    s = 'AS1234 AS 4321 ASN 5678'
    asn_list = list(asns_find(s))
    assert len(asn_list) == 3
    assert 'ASN1234' in asn_list
    assert 'ASN4321' in asn_list
    assert 'ASN5678' in asn_list


@pytest.mark.network
def test_asns_docs_1():
    data = asns()
    data_list = list(data)
    assert len(data_list) >= 62934
    assert isinstance(data_list[0], tuple)
    assert isinstance(data_list[0][0], str)


@pytest.mark.network
def test_asns_private_numbers_docs_1():
    private_numbers = asns_private_numbers()
    data_list = list(private_numbers)
    assert len(data_list) == 94968355
    assert isinstance(data_list[0], int)


@standardize_asn_decorator
def standardize_asn_decorator_test(a):
    """."""
    return a


def test_standardize_asn_decorator_1():
    assert standardize_asn_decorator_test('123') == 'ASN123'
    assert standardize_asn_decorator_test('123') == 'ASN123'
    assert standardize_asn_decorator_test(123) == 'ASN123'



"""Helpful functions for working with ASNs."""

import functools
import re
from typing import Dict, Iterable, List, Optional, Tuple

from d8s_csv import csv_read_as_dict, csv_read_as_list
from d8s_html import html_to_json
from d8s_networking import get

from .asns_temp_utils import enumerate_human_readable_range, stringify_first_arg


@stringify_first_arg
def asn_standardize(as_number: str) -> Optional[str]:
    """Standardize the ASN format."""
    numbers = re.findall('[0-9]{1,10}', as_number)
    if numbers:
        return 'ASN{}'.format(numbers[0])
    else:
        message = f'The given data ({as_number}) cannot be formatted as an ASN.'
        print(message)
        return None


def standardize_asn_decorator(func):
    """Standardize the first argument as an ASN."""

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        first_arg = args[0]
        other_args = args[1:]

        standardized_first_arg = asn_standardize(first_arg)
        return func(standardized_first_arg, *other_args, **kwargs)

    return wrapper


def _cidr_report_org_asn_format(as_number: str) -> str:
    """Return the as_number in the format required by cidr-report.org (e.g. "AS1234")."""
    return as_number.replace('N', '')


@standardize_asn_decorator
def asn_announced_prefixes(as_number: str) -> Iterable[str]:
    """."""
    as_number = _cidr_report_org_asn_format(as_number)

    url = f'https://www.cidr-report.org/cgi-bin/as-report/as-report?as={as_number}&view=2.0&v=4&filter=pass'
    html_repsponse = get(url, process_response=True)
    d = html_to_json(html_repsponse)
    prefix_data = d['html'][0]['body'][0]['pre'][0]['a']
    # the prefix_data variable looks like:
    # [{'_attributes': {'href': 'https://62.220.244.0.22.potaroo.net', 'class': ['black']}, '_value': '62.220.244.0/22'}, ...]  # noqa: E501

    for prefix in prefix_data:
        yield prefix['_value']


@standardize_asn_decorator
def asn_adjacent_asns(as_number: str) -> Iterable[str]:
    """."""
    as_number = _cidr_report_org_asn_format(as_number)

    url = f'https://www.cidr-report.org/cgi-bin/as-report?as={as_number}&view=2.0'
    html_repsponse = get(url, process_response=True)
    d = html_to_json(html_repsponse)
    adjacency_report = d['html'][0]['body'][0]['p'][1]['p'][0]['ul'][0]['p'][0]['pre'][0]
    # the adjacency_report variable looks like:
    # {'a': [{'_attributes': {'href': '/cgi-bin/as-report?as=AS6702&v=4&view=2.0'}, '_value': 'AS6702'}, {'_attributes': {'href': '/cgi-bin/as-report?as=AS3326&v=4&view=2.0'}, '_value': 'AS3326'}, {'_attributes': {'href': '/cgi-bin/as-report?as=AS1&v=4&view=2.0'}, '_value': 'AS1'}], '_values': ['48085 IDATACENTER, CZ\n\n  Adjacency:     3  Upstream:     2  Downstream:     1\n  Upstream Adjacent AS list', 'APEXNCC-AS Gagarina avenue, building 7, room 61, RU', 'DATAGROUP "Datagroup" PJSC, UA\n  Downstream Adjacent AS list', 'LVLT-1, US']}  # noqa: E501

    for link in adjacency_report['a']:
        yield link['_value']


# @standardize_asn_decorator
# def asn_whois(as_number: str) -> str:
#     from d8s_html import website_get_section_containing

#     as_number = _cidr_report_org_asn_format(as_number)

#     url = f'https://www.cidr-report.org/cgi-bin/as-report?as={as_number}&view=2.0'
#     whois_data = website_get_section_containing(url, 'whois: 399260')
#     return whois_data


def asns_find(text: str) -> Iterable[str]:
    """Parse ASNs from the given text."""
    from ioc_finder import ioc_finder

    yield from ioc_finder.parse_asns(text)


def asns() -> Iterable[Tuple[str, str]]:
    """Get a list of ASNs from http://bgp.potaroo.net/as1221/asnames.txt."""
    url = 'http://bgp.potaroo.net/as1221/asnames.txt'

    raw_data = get(url, process_response=True)
    raw_data = re.sub(r'\s(?:\s)+', '\t', raw_data)
    csv_asn_names = csv_read_as_list(raw_data, delimiter='\t')

    for data_point in csv_asn_names:
        asn = asn_standardize(data_point[0])
        asn_temp_name = data_point[1]
        yield asn, asn_temp_name


@standardize_asn_decorator
def asn_number(as_number: str) -> int:
    """Get the number value of the given ASN."""
    return int(as_number.lstrip('ASN'))


def asn_is_private(as_number: str) -> bool:
    """Check if the given ASN is private."""
    private_asn_numbers = asns_private_numbers()
    return asn_number(as_number) in private_asn_numbers


def asns_private_numbers() -> Iterable[int]:
    """Return the reserved (private) ASN numbers.

    Data is collected from:

    https://www.iana.org/assignments/iana-as-numbers-special-registry/iana-as-numbers-special-registry.xhtml

    This function only returns the private ASN numbers.
    The `asns_private_ranges` function returns more information about the private ASN ranges.
    """
    private_asn_data = asns_private_ranges()

    for private_asn_entry in private_asn_data:
        private_asn_numbers = private_asn_entry['AS Number']
        if '-' in private_asn_numbers:
            yield from enumerate_human_readable_range(private_asn_numbers)
        else:
            yield int(private_asn_numbers)


def asns_private_ranges() -> List[Dict[str, str]]:
    """Return the reserved (private) ASN ranges.

    Data is collected from:

    https://www.iana.org/assignments/iana-as-numbers-special-registry/iana-as-numbers-special-registry.xhtml
    """
    private_asns = csv_read_as_dict(
        get(
            'https://www.iana.org/assignments/iana-as-numbers-special-registry/special-purpose-as-numbers.csv',
            process_response=True,
        ),
    )
    return private_asns


@standardize_asn_decorator
def asn_name(as_number: str) -> Optional[str]:
    """Get the name of the given asn."""
    matching_name: Optional[str] = None
    all_asns = asns()

    for asn, name in all_asns:
        if asn == as_number:
            matching_name = name

    return matching_name
